
try:
    from version import __version__
except ImportError:
    __version__ = 'unknown (package not built using setuptools)'

from parameter import *
from twonport import *
from nport import *
import deemb
import tline
import touchstone
import citi



import nport
import numpy as np


class Deembedder(object):
    def __init__(self):
        """
        Set up the de-embedder by passing the nport parameters of the dummy
        structures.
        
        """
        raise NotImplementedError
    
    def deembed(self, measurement):
        """
        De-embed the given measurement parameters

        :param measurement: full measurement structure
        :type measurement: :class:`nport.NPort`
        :returns: de-embedded parameters
        :rtype: :class:`nport.NPort` (S)
        
        """
        raise NotImplementedError


class TwoStep(Deembedder):
    """A simple two-step (open-short) de-embedder"""
    def __init__(self, open, short):
        """
        :param open: open structure
        :type open: :class:`nport.NPort`
        :param short: short structure
        :type short: :class:`nport.NPort`

        """
        y_short = short.convert(nport.Y)
        self.y_open = open.convert(nport.Y)
        self.z_so = (y_short - self.y_open).convert(nport.Z)

    def deembed(self, measurement):
        y_full = measurements.convert(nport.Y)
        z_ = (y_full - self.y_open).convert(nport.Z)
        z_dut = z_ - self.z_so
        
        return z_dut.convert(nport.S)

    deembed.__doc__ = Deembedder.deembed.__doc__


class Vandamme01(Deembedder):
    """
    "Improved Three-Step De-Embedding Method to Accurately Account for the
    Influence of Pad Parasitics in Silicon On-Wafer RF Test-Structures"
    by Ewout P. Vandamme, Dominique M. M.-P. Schreurs, and Cees van Dinther in
    *IEEE Transactions on Electron Devices*, vol. 48, no. 4, pp. 737-742, 2001
    
    """
    def __init__(self, open, short1, short2, through):
        """
        :param open: open structure
        :type open: :class:`nport.NPort`
        :param short1: port 1 short structure
        :type short1: :class:`nport.NPort`
        :param short2: port 2 short structure
        :type short2: :class:`nport.NPort`
        :param through: through structure
        :type through: :class:`nport.NPort`

        """
        # convert S to Y parameters
        y_open = open.convert(nport.Y)
        y_short1 = short1.convert(nport.Y)
        y_short2 = short2.convert(nport.Y)
        y_through = through.convert(nport.Y)

        # extract parameters from dummy structures
        y11_open = y_open.get_element(1, 1)
        y12_open = y_open.get_element(1, 2)
        y22_open = y_open.get_element(2, 2)
        y11_short1 = y_short1.get_element(1, 1)
        y22_short2 = y_short2.get_element(2, 2)
        y12_through = y_through.get_element(1, 2)

        self.g1 = y11_open + y12_open
        self.g2 = y22_open + y12_open
        self.g3 = (- y12_open.convert(nport.Z) +
                     y12_through.convert(nport.Z)).convert(nport.Y)

        z_x = y12_through.convert(nport.Z)
        z_y = (y11_short1 - self.g1).convert(nport.Z)
        z_z = (y22_short2 - self.g2).convert(nport.Z)

        self.z1 = 0.5 * (- z_x + z_y - z_z)
        self.z2 = 0.5 * (- z_x - z_y + z_z)
        self.z3 = 0.5 * (+ z_x + z_y + z_z)

    def deembed(self, measurement):
        y_meas = measurement.convert(nport.Y)
        y_a = y_meas - (np.asarray([[1, 0], [0, 0]]) * self.g1 +
                        np.asarray([[0, 0], [0, 1]]) * self.g2)
        z_a = y_a.convert(nport.Z)
        z_b = z_a - (np.asarray([[1., 0], [0, 0]]) * self.z1 +
                     np.asarray([[0, 0], [0, 1.]]) * self.z2 +
                     np.ones((2,2)) * self.z3)
        y_b = z_b.convert(nport.Y)
        y_dut = y_b - (np.asarray([[1., -1.], [-1., 1.]]) * self.g3)
        
        return y_dut.convert(nport.S)

    deembed.__doc__ = Deembedder.deembed.__doc__


kol00_asym = True

class Kolding00(Deembedder):
    """
    "A Four-Step Method for De-Embedding Gigahertz On-Wafer CMOS
    Measurements" by Troels Emil Kolding in *IEEE Transactions on Electron 
    Devices*, vol. 47, no. 4, pp. 734-740, 2000
    
    """
    def __init__(self, simple_open, simple_short, open, short1, short2, alpha=0.0):
        """        
        :param simpleopen: simple open structure
        :type simpleopen: :class:`nport.NPort`
        :param simpleshort: simple short structure
        :type simpleshort: :class:`nport.NPort`
        :param open: open structure
        :type open: :class:`nport.NPort`
        :param short1: port 1 short structure
        :type short1: :class:`nport.NPort`
        :param short2: port 2 short structure
        :type short2: :class:`nport.NPort`
        :param through: through structure
        :type through: :class:`nport.NPort`
        :param alpha: compensation parameter to account for the imperfections
                      of the short standard when gaps become large
        :type alpha: float

        """
        # convert S to Z parameters
        z_simpleopen = simple_open.convert(nport.Z)
        z_simpleshort = simple_short.convert(nport.Z)
        z_open = open.convert(nport.Z)
        z_short1 = short1.convert(nport.Z)
        z_short2 = short2.convert(nport.Z)

        # extract parameters from dummy structures
        z11_ss = z_simpleshort.get_element(1, 1)
        z22_ss = z_simpleshort.get_element(2, 2)
        z11_so = z_simpleopen.get_element(1, 1)
        z22_so = z_simpleopen.get_element(2, 2)

        # pads
        if kol00_asym:
            self.zc = 2.0/3.0 * z_simpleshort
            self.zp = (z_simpleopen - z_simpleshort)
        else:
            self.zc = 2.0/3.0 * np.identity(2) * z_simpleshort
            self.zp = np.identity(2) * (z_simpleopen - z_simpleshort)
        
        # remove pads from short1 and open
        z_short1__ = self._remove_pads(z_short1)
        z_short2__ = self._remove_pads(z_short2)
        z_open__ = self._remove_pads(z_open)

        z11_s1__ = z_short1__.get_element(1, 1)
        z21_s1__ = z_short1__.get_element(2, 1)
        z12_s1__ = z_short1__.get_element(1, 2)
        z22_s2__ = z_short2__.get_element(2, 2)
        z12_s2__ = z_short2__.get_element(1, 2)
        z21_s2__ = z_short2__.get_element(2, 1)
        z11_o__ = z_open__.get_element(1, 1)
        z21_o__ = z_open__.get_element(2, 1)
        z12_o__ = z_open__.get_element(1, 2)
        z22_o__ = z_open__.get_element(2, 2)

        # calculate remaining parameters
        if kol00_asym:
            z21 = 0.5 * (z21_s1__ + z12_s1__)
            z22 = 0.5 * (z21_s2__ + z12_s2__)
            self.z2 = np.ones((2, 2)) * (z21 + z22) * 0.5
            zi1_plus_z11 = (z11_s1__ - z21) / (1.0 + alpha)
            zi2_plus_z12 = (z22_s2__ - z22) / (1.0 + alpha)
            self.zi_plus_z1 = np.asarray([[1, 0], [0, 0]]) * zi1_plus_z11 + \
                              np.asarray([[0, 0], [0, 1]]) * zi2_plus_z12
            z31 = z21_o__ + z11_o__ - 2.0 * z21 - zi1_plus_z11
            z32 = z12_o__ + z22_o__ - 2.0 * z22 - zi2_plus_z12
            self.y3 = np.asarray([[1, 0], [0, 0]]) * z31.convert(nport.Y) + \
                      np.asarray([[0, 0], [0, 1]]) * z32.convert(nport.Y)
            zf1 = z31 * (z31 / (z21_o__ - z21) - 2)
            zf2 = z32 * (z32 / (z12_o__ - z22) - 2)
            #self.yf = np.asarray([[0, 1], [0, 0]]) * zf1.convert(nport.Y) + \
            #          np.asarray([[0, 0], [1, 0]]) * zf2.convert(nport.Y)
            self.yf = 0.5 * np.asarray([[0, 1], [1, 0]]) * \
                      (zf1 + zf2).convert(nport.Y)
        else:
            self.z2 = 0.5 * (z21_s1__ + z12_s1__)
            self.zi_plus_z1 = (z11_s1__ - self.z2) / (1.0 + alpha)
            z3 = z21_o__ + z11_o__ - 2.0 * self.z2 - self.zi_plus_z1
            zf = z3 * (z3 / (z21o__ - self.z2) - 2.0)
            self.y3 = z3.convert(nport.Y)
            self.yf = Zf.convert(nport.Y)

    def deembed(self, measurement):
        z_full = measurement.convert(nport.Z)
        z__ = self._remove_pads(z_full)

        if kol00_asym:
            print("WARNING: de-embedding method adjusted for asymmetrical fixtures")
            print("WARNING: might NOT be CORRECT")
            z___ = z__ - (self.zi_plus_z1 + self.z2)
            y___ = z___.convert(nport.Y)
            y_dut = y___ - (self.y3 - self.yf)
        else:
            z___ = z__ - (self.zi_plus_z1 * np.identity(2) +
                          self.z2 * np.ones((2,2)))
            y___ = z___.convert(nport.Y)
            y_dut = y___ - (self.y3 * np.identity(2) +
                            - self.yf * np.asarray([[0,1],[1,0]]))
        
        return y_dut.convert(nport.S)

    deembed.__doc__ = Deembedder.deembed.__doc__

    def _remove_pads(self, z_structure):
        """De-embed the pads from z_structure
        
        :param z_structure: structure to remove pads from
        :type z_structure: :class:`nport.NPort` (Z)
        :returns: structure with pads removed
        :rtype: :class:`nport.NPort` (Z)
        
        """
        assert z_structure.type == nport.Z
        z_ = z_structure - (3.0/2.0 * self.zc)
        y_ = z_.convert(nport.Y)

        y_nopads = y_ - self.zp.convert(nport.Y)
        z_nopads = y_nopads.convert(nport.Z)
        return z_nopads




import re
import os
from datetime import datetime
import numpy as np
import nport


def read(file_path, verbose=False):
    """
    Load the contents of a CITI file into an NPort
    
    :returns: NPort holding data contained in the CITI file
    :rtype: :class:`nport.NPort`
    
    """
    file_path = os.path.abspath(file_path)
    citifile = CITIFile(file_path)
    assert citifile.params[0][0][0] == "freq"
    freqs = citifile.data[0][0]
    ports = np.sqrt(len(citifile.params[0]) - 1)
    assert ports == int(ports)
    ports = int(ports)
    
    re_param = re.compile(r"^S\[(\d+),(\d+)\]$")
    indices = []
    for param in citifile.params[0][1:]:
        name = param[0]
        m = re_param.match(name)
        port1 = int(m.group(1))
        port2 = int(m.group(2))
        indices.append((port1, port2))

    matrices = []
    for index in range(len(freqs)):
        matrix = np.array([[None for i in range(ports)]
            for j in range(ports)], dtype=complex)
        for i, port in enumerate(indices):
            port1 = port[0]
            port2 = port[1]
            matrix[port1 - 1, port2 - 1] = citifile.data[0][i+1][index]
        matrices.append(matrix)
    return nport.NPort(freqs, matrices, nport.SCATTERING, 50)


def write(instance, file_path):
    """Write the n-port data held in `instance` to a CITI file at file_path.
    
    :param instance: n-port data
    :type instance: :class:`nport.NPort`
    :param file_path: filename to write to (without extension)
    :type file_path: str
    
    """
    file_path = file_path + ".citi"
    file = open(file_path, 'wb')
    creationtime = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    file.write("# Created by the Python nport module\n")
    file.write("# Creation time: %s\n" % creationtime)
    file.write("CITIFILE A.01.01\n")
    file.write("VAR freq MAG %d\n" % len(instance.freqs))
    instance = instance.convert(nport.S, 50)
    for i in range(instance.ports):
        for j in range(instance.ports):
            file.write("DATA S[%d,%d] RI\n" % (i + 1, j + 1))
    file.write("VAR_LIST_BEGIN\n")
    for freq in instance.freqs:
        file.write("\t%g\n" % freq)
    file.write("VAR_LIST_END\n")
    for i in range(instance.ports):
        for j in range(instance.ports):
            file.write("BEGIN\n")
            for parameter in instance.get_parameter(i + 1, j + 1):
                file.write("\t%g, %g\n" % (parameter.real, parameter.imag))
            file.write("END\n")
    file.write("\n")


# Collection of object classes for reading calibration lab data file types
#
# Author:           J. Wayde Allen
# Creation Date:    2001-05-22
# Revised:          2001-05-23 JWA
#                   2010-01-28 Brecht Machiels
#                               * made parsing more robust
#                               * changed indentation from 3 to 4 spaces

import string, sys

class CITIFile:
    def __init__(self, filename):
        self.filename = filename

        # The following are the main data structures
        self.packages = {}        
        self.constants = []
        self.params = []
        self.data = []
        self.instrmnt = []

        # Open the citifile 
        myfile = open(self.filename, 'r')

        # Define some special control and book keeping variables
        packagecounter = -1 # Index to the number of Citifile packages
        packagenames = []   # List of the package names

        while 1:
            line = myfile.readline()

            if not line:
                break
            
            linetxt = string.strip(line)
            line = string.split(linetxt)

            #This line starts a new Citifile data package
            #update the package counter and create blank indices
            if len(line) > 0:

                if line[0] == 'CITIFILE':
                    packagecounter = packagecounter + 1
                    packagenames.append("")  #Create a blank name entry

                    self.constants.append([])
                    self.params.append([])
                    self.data.append([])
                    self.instrmnt.append([])

                    indata = 'NO'       #Not reading data
                    invarlist = 'NO'    #Not reading independant variable data
                    datacount = 0       #Index to package data blocks

                #Skip device-specific variables
                if line[0][0] == '#':
                    continue

                #Should be one name per package
                elif line[0] == 'NAME':  
                    packagenames[packagecounter] = line[1]

                elif line[0] == 'CONSTANT':
                    self.constants[packagecounter].append((line[1],line[2]))

                elif line[0] == 'VAR':
                    self.params[packagecounter].append((line[1],line[2],line[3]))

                elif line[0] == 'SEG_LIST_BEGIN':
                    invarlist = 'SEG'
                    self.data[packagecounter].append([])

                elif line[0] == 'SEG' and invarlist == 'SEG':

                    #Decode the start, stop and number of points entries
                    start = float(line[1])
                    stop  = float(line[2])
                    numpoints  = int(line[3])

                    #Compute the actual data values from this information
                    #and put it in the data block
                    step = (stop - start) / (numpoints - 1)
                    next = start
                    count = 0
                    while next <= stop:
                        count = count + 1
                        self.data[packagecounter][datacount].append(next)
                        next = next + step

                elif line[0] == 'SEG_LIST_END':
                    invarlist = 'NO'
                    #We've filled this data bin so point to the next one
                    datacount = datacount + 1

                elif line[0] == 'VAR_LIST_BEGIN':
                    invarlist = 'VARLIST'
                    self.data[packagecounter].append([])

                elif line[0] != 'VAR_LIST_END' and invarlist == 'VARLIST':
                    datum = float(line[0])
                    self.data[packagecounter][datacount].append(datum)

                elif line[0] == 'VAR_LIST_END':
                    invarlist = 'NO'
                    datacount = datacount + 1

                elif line[0] == 'DATA':
                    self.params[packagecounter].append((line[1],line[2]))

                elif line[0] == 'BEGIN': 
                    indata = 'YES'
                    self.data[packagecounter].append([])

                elif line[0] != 'END' and indata == 'YES':

                    if self.params[packagecounter][datacount][1] == 'RI':
                        real,imag = string.split(linetxt,',')
                        value = complex(float(real),float(imag))

                    elif self.params[packagecounter][datacount][1] == 'MAG':
                        value = float(line[0])

                    self.data[packagecounter][datacount].append(value)

                elif line[0] == 'END': 
                    indata = 'NO'
                    datacount = datacount + 1

                else:
                    #Anything else must be instrument specific so make these
                    #lines available for parsing by the user
                    self.instrmnt[packagecounter].append(line)

        #We've read and sorted all of these data
        #Create dictionary of package index and names
        for values in range(0,packagecounter+1):
            self.packages[values] = packagenames[values]



import re
import os
from datetime import datetime
import numpy as np
import nport
import parameter

REAL_IMAG = 'RI'
MAG_ANGLE = 'MA'
DB_ANGLE = 'DB'

keys = {}
keys[REAL_IMAG] = ('real', 'imag')
keys[MAG_ANGLE] = ('mag', 'deg')
keys[DB_ANGLE] = ('db', 'deg')


def read(file_path, verbose=False):
    """
    Load the contents of a Touchstone file into an NPort
    
    :returns: NPort holding data contained in the Touchstone file
    :rtype: :class:`nport.NPort`
    
    """
    re_filename = re.compile(r"\.s(?P<ports>\d+)+p", re.I)
    m = re_filename.search(file_path)
    ports = int(m.group('ports'))
    if verbose:
        print("File '%s'" % file_path)
        print("  Number of ports (based on file extension) = %d" % \
            ports)
    file_path = os.path.abspath(file_path)
    file = open(file_path, 'rb')
    (frequnit, type, format, z0) = _parse_option_line(file, verbose)
    freqs = []
    matrices = []
    try:
        while True:
            freq, matrix = _parse_next_sample(file, ports, format)
            freqs.append(freq)
            matrices.append(matrix)
    except EOFError:
        pass
    finally:
        file.close()
    return nport.NPort(np.asarray(freqs) * frequnit, matrices, type, z0)


_re_comment = re.compile(r"^\s*!")
_re_options = re.compile(r"^\s*#")
_re_empty   = re.compile(r"^\s*$")


def _parse_option_line(file, verbose=False):
    """Parse and interpret the option line in the touchstone file"""
    regex = {
        'unit':      re.compile(r"(#|\s+)(?P<unit>G|M|K|)Hz(\s+|$)", re.I),
        'parameter': re.compile(r"(#|\s+)(?P<parameter>S|Y|Z|H|G)(\s+|$)", re.I),
        'format':    re.compile(r"(#|\s+)(?P<format>DB|MA|RI)(\s+|$)", re.I),
        'reference': re.compile(r"(#|\s+)R\s+(?P<reference>\d+)(\s+|$)", re.I)
    }
    
    prefix = {
        '' : 1,
        'K': 1e3,
        'M': 1e6,
        'G': 1e9
    }

    # defaults
    frequnit = 1e9
    type = nport.SCATTERING
    format = MAG_ANGLE
    z0 = 50
    
    # format of the options line (order is unimportant)
    # <frequency unit> <parameter> <format> R <n>
    line = file.readline()
    while not line.startswith('#'):
        line = file.readline()

    m = regex['unit'].search(line)
    if m:
        frequnit = prefix[m.group('unit').upper()]
    m = regex['parameter'].search(line)
    if m:
        type = m.group('parameter').upper()
    m = regex['format'].search(line)
    if m:
        format = m.group('format').upper()
    m = regex['reference'].search(line)
    if m:
        z0 = float(m.group('reference'))
           
    if verbose:
        print("  Frequency unit: %g Hz" % frequnit)
        print("  Parameter:      %s"    % type)
        print("  Format:         %s"    % format)
        print("  Reference R:    %g"    % z0)
        
    # only S-parameters are supported for now
    if type != nport.SCATTERING:
        raise NotImplementedError
    
    return (frequnit, type, format, z0)


def _get_next_line_data(file):
    """
    Returns the data on the next line of the input file as an array 
    of floats, skipping comment, option and empty lines.
    
    """
    line = '!'
    while _re_comment.search(line) or \
        _re_options.search(line) or \
        _re_empty.search(line):
        line = file.readline()
        if not line: # end of data
            raise EOFError
    data = []
    line = line.split('!')[0]
    for number in line.split():
        data.append(float(number))
    return data


def _parse_next_sample(file, ports, format):
    """Parse the parameters for the next frequency point"""
    data = _get_next_line_data(file)
    # data lines always contain an even number of values (2 values
    #  per parameter), *unless* a frequency value is included in
    #  front
    assert len(data) % 2 == 1
    freq = data[0]
    data = data[1:]
    count = 0

    port1 = port2 = 1
    matrix = np.array([[None for i in range(ports)]
        for j in range(ports)], dtype=complex)

    while True:
        for i in range(len(data) / 2):
            index = 2 * i
            args = {}
            args[keys[format][0]] = data[index]
            args[keys[format][1]] = data[index + 1]
            try:
                if ports == 2:			
                    matrix[port1 - 1, port2 - 1] = parameter.parameter(**args)
                else:			
                    matrix[port2 - 1, port1 - 1] = parameter.parameter(**args)
            except IndexError:
                raise Exception("more ports than reported in the file "
                                "extension")

            port1 += 1
            if port1 > ports:
                port2 += 1
                port1 = 1                    

        count += len(data) / 2
        if port2 > ports:
            break
        data = _get_next_line_data(file)
        if len(data) % 2 != 0:
            raise Exception("less ports than reported in the file extension")

    assert count == ports**2
    return (freq, matrix)


def write(instance, file_path, format=REAL_IMAG):
    """Write the n-port data held in `instance` to a Touchstone file at
    file_path.
    
    :param instance: n-port data
    :type instance: :class:`nport.NPort`
    :param file_path: filename to write to (without extension)
    :type file_path: str
    :param format: determines format of the Touchstone file ('MA', 'RI' or 'DB')
    :type format: str
  
    """
    file_path = file_path + ".s%dp" %  instance.ports
    file = open(file_path, 'wb')
    creationtime = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    file.write("! Created by the Python nport module\n")
    file.write("! Creation time: %s\n" % creationtime)
    file.write("# Hz %s %s R %g\n" % (instance.type, format, instance.z0))
    flatten_order = 'F' if instance.ports == 2 else 'C'
    threeport = (instance.ports == 3)
    for i, freq in enumerate(instance.freqs):
        sample = "\t%g\t" % freq
        parameters = instance[i].flatten(flatten_order)
        for i, element in enumerate(parameters):
            if i != 0 and \
                ((threeport and i % 3 == 0) or (not threeport and i % 4 == 0)):
                sample += "\n\t\t"
            if format == REAL_IMAG:
                sample += " %g %g" % (parameter.real(element), parameter.imag(element))
            elif format == MAG_ANGLE:
                sample += " %g %g" % (parameter.mag(element), parameter.deg(element))
            elif format == DB_ANGLE:
                sample += " %g %g" % (parameter.db(element), parameter.deg(element))
            else:
                raise ValueError("unknown format specified")
        file.write(sample + "\n")




from __future__ import division

import numpy as np
from scipy import interpolate
import parameter
import twonport


# parameter matrix types
Z = IMPEDANCE = TYPE_Z = 'Z'
Y = ADMITTANCE = TYPE_Y = 'Y'
S = SCATTERING = TYPE_S = 'S'
T = SCATTERING_TRANSFER = TYPE_T = 'T'
H = HYBRID = TYPE_H = 'H'
G = INVERSE_HYBRID = TYPE_G = 'G'
ABCD = TRANSMISSION = TYPE_ABCD = 'ABCD'


#~ class MatrixTypeError(Exception):
    #~ """Inappropriate parameter matrix type"""
    #~ def __init__(self, given, expected):
        #~ self.given = given
        #~ self.expected = expected
    #~ def __str__(self):
        #~ return "given type %s does not match expected type %s" % (self.given, self.expected)



class NPortMatrixBase(np.ndarray):
    """Base class representing an n-port matrix (Z, Y, S, T, G, H or ABCD)"""
    # TODO: implement type checking for operations (as in NPortBase)
    def __new__(cls, matrix, mtype, z0=None):
        # http://docs.scipy.org/doc/numpy/user/basics.subclassing.html
        obj = np.asarray(matrix, dtype=complex).view(cls)
        if mtype not in (Z, Y, S, T, H, G, ABCD):
            raise ValueError("illegal matrix type specified")
        obj.type = mtype
        obj.z0 = NPortMatrixBase.__z0test__(mtype, z0)
        return obj

    @classmethod
    def __z0test__(cls, mtype, z0):
        if mtype in (S, T):
            if z0 is not None:
                return z0
            else:
                return 50.0
        else:
            if z0 is not None:
                raise ValueError("the specified n-port representation (%s) does"
                                 " not require a reference impedance" % mtype)
            else:
                return None

    def __array_finalize__(self, obj):
        if obj is None: return
        self.type = getattr(obj, 'type', None)
        self.z0 = getattr(obj, 'z0', None)

    @property
    def ports(self):
        raise NotImplementedError


class NPortBase(NPortMatrixBase):
    """Base class representing an n-port across a list of frequencies"""
    def __metaclass__(cname, cbases, cdict):
        # http://thread.gmane.org/gmane.comp.python.general/373788/focus=373795
        P = type(cname, cbases, cdict)
        def make_op(name, func):
            def op(this, other):
                if issubclass(type(other), NPortBase):
                    if other.type == this.type:
                        if other.z0 != this.z0:
                            raise ValueError("operands have different Z0")
                        result_type = this.type
                        result_z0 = this.z0
                    else:
                        raise ValueError("operands have different types")
                    min_freq = max(this.freqs[0], other.freqs[0])
                    max_freq = min(this.freqs[-1], other.freqs[-1])
                    result_freqs = list(set(this.freqs).union(set(other.freqs)))
                    result_freqs.sort()
                    result_freqs = \
                        np.array(result_freqs[result_freqs.index(min_freq):
                                              result_freqs.index(max_freq)+1])
                    this_matrices = this.at(result_freqs)
                    other_matrices = other.at(result_freqs)
                    result_matrices = func(this_matrices, other_matrices)
                else:
                    result_freqs = this.freqs
                    result_matrices = func(this, other)
                    result_type = this.type
                    result_z0 = this.z0
                subclass = type(this)
                return subclass(result_freqs, result_matrices, result_type,
                                result_z0)
            op.__name__ = name
            return op
        for name, func in np.ndarray.__dict__.items():
            if callable(func) and name in ['__add__', '__sub__', '__mul__',
                '__div__', '__radd__', '__rsub__', '__rmul__', '__rdiv__']:
                setattr(P, name, make_op(name, func))
        return P

    def __new__(cls, freqs, matrices, mtype, z0=None):
        if len(freqs) != len(matrices):
            raise ValueError("the list of frequencies and the list of "
                             "matrices should have equal lenghts")
        if matrices[0].shape[0] != matrices[0].shape[1]:
            raise ValueError("the matrices should be square")
        obj = NPortMatrixBase.__new__(cls, matrices, mtype, z0)
        obj.freqs = np.asarray(freqs)
        return obj

    def __array_finalize__(self, obj):
        NPortMatrixBase.__array_finalize__(self, obj)
        self.freqs = getattr(obj, 'freqs', None)

    def __getitem__(self, arg):
        if type(arg) == int:
            return NPortMatrix(np.asarray(self).__getitem__(arg), self.type,
                               self.z0)
        else:
            return np.asarray(self).__getitem__(arg)

    #~ def __repr__(self):
        #~ return "freqs: " + repr(self.freqs) + "\n" + NPortMatrixBase.__repr__(self)

    #~ def __str__(self):
        #~ return self.__repr__()

    def get_parameter(self, port1, port2):
        """Return the parameter as specified by the indices port1 and port2 as
        an ndarray
        
        """
        return np.asarray(self[:, port1 - 1, port2 - 1])

    def get_element(self, port1, port2):
        """Return the submatrices made up of the element as specified by the
        indices port1 and port2
        
        """
        subclass = type(self)
        return subclass(self.freqs,
            np.asarray(self[:, (port1 - 1, ), :][:, :, (port2 - 1, )]),
            self.type, self.z0)

    def at(self, freqs):
        """Return the interpolated n-port data at the given
        * list of frequencies (freqs is iterable), or
        * at a single frequency (freqs is a value)
        
        """
        # check whether freqs is iterable
        try:
            it = iter(freqs)
            single = False
        except TypeError:
            single = True
            freqs = [freqs]
        subclass = type(self)
        func = interpolate.interp1d(self.freqs, self, axis=0)
        interpolated = func(freqs)
        interpolated_nport = subclass(freqs, interpolated, self.type, self.z0)
        if single:
            return interpolated_nport[0]
        else:
            return interpolated_nport
            
    def average(self, n):
        """Take a moving average over n frequency samples"""
        averaged = np.zeros_like(np.asarray(self))
        for i in range(len(self)):
            for j in range(- int(np.floor(n/2)), int(np.ceil(n/2))):
                index = i + j
                if index < 0:
                    index = 0
                elif index >= len(self):
                    index = len(self) - 1
                averaged[i] += self[index] / n        
        # numpy convolve
        #~ ones = np.ones(n) / n
        #~ averaged = np.zeros_like(np.asarray(self))
        #~ for i in range(self.ports):
            #~ for j in range(self.ports):
                #~ averaged[:, i, j] = np.convolve(ones, self[:, i, j])
        # or scipy convolve
        #~ averaged = scipy.signal.convolve(np.ones((n, ) + self.shape[1:])/n, np.asarray(self), 'same')
        subclass = type(self)
        return subclass(self.freqs, averaged, self.type, self.z0)


class NPortMatrix(NPortMatrixBase):
    """Class representing an n-port matrix (Z, Y, S, T, G, H or ABCD)"""
    def __new__(cls, matrix, mtype, z0=None):
        obj = NPortMatrixBase.__new__(cls, matrix, mtype, z0)
        if len(obj.shape) != 2:
            raise ValueError("the matrix should be two-dimensional")
        if obj.shape[0] != obj.shape[1]:
            raise ValueError("the matrix should be square")
        return obj

    @property
    def ports(self):
        return self.shape[0]

    def twonportmatrix(self, inports=None, outports=None):
        """Return the 2n-port matrix represented by this n-port matrix
        
        :param inports: the list of ports that make up the inputs of the 2n-port
        :type inports: tuple or list
        :param outports: the list of ports that make up the outputs
        :type outports: tuple or list
        :rtype: :class:`TwoNPortMatrix`
        
        """
        if self.ports % 2 != 0:
            raise TypeError("this NPortMatrix' number of ports is not a "
                            "multiple of 2")
        n = int(self.ports / 2)
        if inports is None and outports is None:
            inports = range(n)
            outports = [i + n for i in range(n)]
            matrix = self
        else:
            # check whether the given sets of ports are valid
            assert inports is not None and outports is not None
            assert len(inports) == n
            assert len(outports) == n
            allports = set(inports).union(set(outports))
            assert len(allports) == 2*n
            assert min(allports) == 1 and max(allports) == 2*n
            inports = [port - 1 for port in inports]
            outports = [port - 1 for port in outports]

            ports = inports + outports

            # shuffle rows and columns to obtain 2n-port format
            matrix = []
            for row in ports:
                matrix.append(self[row])
            matrix = np.asarray(matrix).T
            matrix2 = []
            for column in ports:
                matrix2.append(matrix[column])
            matrix = np.asarray(matrix2).T

        matrix = np.asmatrix(matrix)
        x11 = matrix[0:n  , 0:n  ]
        x12 = matrix[0:n  , n:2*n]
        x21 = matrix[n:2*n, 0:n  ]
        x22 = matrix[n:2*n, n:2*n]
        matrix = np.array([[x11, x12], [x21, x22]])
        return twonport.TwoNPortMatrix(matrix, self.type, self.z0)

    def convert(self, mtype, z0=None):
        """Convert to another n-port matrix representation"""
        # references:
        #  * http://qucs.sourceforge.net/tech/node98.html
        #           "Transformations of n-Port matrices"
        #  * Multiport S-Parameter Measurements of Linear Circuits With Open Ports
        #           by Reimann et al. (only S <-> Z)
        #  * Agilent AN 154 - S-Parameter Design (S <-> T)
        #  * MATLAB S-parameter toolbox (Z, Y, H, G, ABCD, T)
        #           http://www.mathworks.com/matlabcentral/fileexchange/6080
        z0 = NPortMatrixBase.__z0test__(mtype, z0)
        identity = np.identity(len(self), dtype=complex)

        # TODO: check for singularities
        if self.type == SCATTERING:
            if mtype == IMPEDANCE:
                result = self.z0 * (2 * np.linalg.inv(identity - self) -
                                    identity)
            elif mtype == ADMITTANCE:
                # TODO: need to verify
                result = 1 / self.z0 * np.dot(np.linalg.inv(identity + self),
                                              identity - self)
            #~ elif mtype == SCATTERING_TRANSFER:
                #~ # only 2-port
                #~ # NOTE: S21 is NOT used - verify!
                #~ assert self.ports == 2
                #~ T11 = 1 / self[0,1]
                #~ T12 = - self[1,1] / self[0,1]
                #~ T21 = self[0,0] / self[0,1]
                #~ T22 = self[0,1] - self[0,0] * self[1,1] / self[0,1]
                #~ result = np.array([[T11, T12], [T21, T22]])
            #~ elif mtype == TRANSMISSION:
                #~ # TODO: only provide in TwoNPortMatrix?
                #~ # only 2-port
                #~ assert self.ports == 2
                #~ denom = 2 * self[1,0]
                #~ A =               ((1 + self[0,0])*(1 - self[1,1]) + self[0,1] * self[1,0]) / denom
                #~ B =     self.z0 * ((1 + self[0,0])*(1 + self[1,1]) - self[0,1] * self[1,0]) / denom
                #~ C = 1.0/self.z0 * ((1 - self[0,0])*(1 - self[1,1]) - self[0,1] * self[1,0]) / denom
                #~ D =               ((1 - self[0,0])*(1 + self[1,1]) + self[0,1] * self[1,0]) / denom
                #~ result = np.array([[A, B], [C, D]])
            elif mtype == SCATTERING:
                if z0 == self.z0:
                    return self
                else:
                    # TODO: renormalize
                    raise NotImplementedError
            else:
                raise NotImplementedError
        elif self.type == IMPEDANCE:
            if mtype == SCATTERING:
                # TODO: renormalize
                result = identity - 2 * np.linalg.inv(identity + self / z0)
            elif mtype == ADMITTANCE:
                result = np.linalg.inv(self)
            elif mtype == IMPEDANCE:
                raise NotImplementedError
                result = self # renormalize
                z0 = self.z0
            else:
                NotImplementedError
        elif self.type == ADMITTANCE:
            if mtype == SCATTERING:
                # TODO: need to verify
                result = np.dot(identity - (z0 * self), 
                                np.linalg.inv(identity + (z0 * self)))
            elif mtype == IMPEDANCE:
                result = np.linalg.inv(self)
            elif mtype == ADMITTANCE:
                raise NotImplementedError
                result = self # renormalize
                z0 = self.z0
            else:
                raise NotImplementedError
        #~ elif self.type == TRANSMISSION:
            #~ if mtype == SCATTERING:
                #~ # TODO: need to verify
                #~ assert self.ports == 2
                #~ A = self[0,0]
                #~ B = self[0,1]
                #~ C = self[1,0]
                #~ D = self[1,1]
                #~ B0 = B / self.z0
                #~ C0 = C * self.z0
                #~ denom = A + B0 + C0 + D
                #~ S11 = (A + B0 - C0 - D) / denom
                #~ S12 = 2 * (A * D - B * C) / denom
                #~ S21 = 2 / denom
                #~ S22 = (-A + B0 - C0 + D) / denom
                #~ result = np.array([[S11, S12], [S21, S22]])
            #~ else:
                #~ raise NotImplementedError
        else:
            raise NotImplementedError

        return NPortMatrix(result, mtype, z0)

    def submatrix(self, ports):
        """Keep only the parameters corresponding to the given ports,
        discarding the others.

        """
        indices = [port - 1 for port in ports]
        submatrix = self[indices, :][:, indices]
        return NPortMatrix(submatrix, self.type, self.z0)

    def recombine(self, portsets):
        """Recombine ports, reducing the number of ports of this NPortMatrix.
        This NPortMatrix has to be an impedance matrix.
        
        `portsets` is a list of ints and tuples. An int specifies the number of
        a port that needs to be kept as-is. If the int is negative, the port's
        polarity will be reversed. A tuple specifies a pair of ports that are
        to be recombined into one port. The second element of this tuple acts
        as the ground reference to the first element.
        
        >>> recombine([(1,3), (2,4), 5, -6]
        
        will generate a four-port where:
        
        * port 1 is original port 1 referenced to port 3
        * port 2 is original port 2 referenced to port 4
        * port 3 is original port 5
        * port 4 is original port 6, but with reversed polarity

        """
        assert self.type == IMPEDANCE
        m = []
        for i, ports in enumerate(portsets):
            row = [0 for port in range(self.ports)]
            try:
                if isinstance(ports, tuple):
                    assert len(ports) == 2
                    row[ports[0] - 1] = 1
                    row[ports[1] - 1] = -1
                else:
                    assert isinstance(ports, int)
                    assert ports != 0
                    if ports > 0:
                        row[ports - 1] = 1
                    else:
                        row[-ports - 1] = -1
            except IndexError:
                raise IndexError("specified port number is higher than number "
                                 "of ports")
            m.append(row)
        m = np.matrix(m, dtype=float)
        result = m * np.asarray(self) * m.T
        return NPortMatrix(result, self.type, self.z0)
        
    def ispassive(self):
        """Check whether this n-port matrix is passive"""
        if self.type != SCATTERING:
            return self.convert(SCATTERING).ispassive()
        else:
            if np.max(np.sum(np.abs(np.asarray(self))**2, 1)) > 1:
                return False
            else:
                return True


class NPort(NPortBase):
    """Class representing an n-port across a list of frequencies"""
    def __new__(cls, freqs, matrices, mtype, z0=None):
        obj = NPortBase.__new__(cls, freqs, matrices, mtype, z0)
        if len(obj[0].shape) != 2:
            raise ValueError("the matrices should be two-dimensional")
        return obj

    def __init__(self, freqs, matrices, type, z0=50):
        """Initialize an instance, specifying the frequency samples, all
        corresponsing matrices, the matrix type and the reference impedance

        """
        pass

    @property
    def ports(self):
        """Return the number of ports of this NPort"""
        return self[0].shape[0]

    def add(self, freq, matrix):
        """Return an NPort with the specified frequency sample added

        If matrix is an NPortMatrix, its elements will be converted to this
        NPort's type and characteristic impedance.
        If matrix is a complex array, it is assumed the elements are in the
        format of this NPort

        """
        if type(matrix) == NPortMatrix:
            if matrix.type != self.type or matrix.z0 != self.z0:
                matrix = matrix.convert(self.type, self.z0)
        index = self.freqs.searchsorted(freq)
        freqs = np.insert(self.freqs, index, freq)
        matrices = np.insert(self.matrices, index, matrix, 0)
        return self.__class__(freqs, matrices, self.type, self.z0)

    def twonport(self, inports=None, outports=None):
        """Convert this NPort to a TwoNPort using inports as the input ports
        and outports as the output ports.

        """
        twonportmatrices = []
        for matrix in self:
            nportmatrix = NPortMatrix(matrix, self.type, self.z0)
            twonportmatrices.append(nportmatrix.twonportmatrix(inports,
                                                               outports))
        return twonport.TwoNPort(self.freqs, twonportmatrices, self.type,
                                 self.z0)

    def convert(self, mtype, z0=None):
        """Convert to another n-port matrix representation"""
        converted = []
        for matrix in self:
            nportmatrix = NPortMatrix(matrix, self.type, self.z0)
            converted.append(nportmatrix.convert(mtype, z0))
        return NPort(self.freqs, converted, mtype, z0)

    def submatrix(self, ports):
        """Keep only the parameters corresponding to the given ports,
        discarding the others.
        """
        indices = [port - 1 for port in ports]
        submatrices = self[:, indices, :][:, :, indices]
        return NPort(self.freqs, submatrices, self.type, self.z0)

    def invert(self):
        """Return an NPort described by the inverse of this NPort's matrices"""
        # TODO: determine type of output
        inverted = []
        for matrix in self:
            nportmatrix = NPortMatrix(matrix, self.type, self.z0)
            inverted.append(np.linalg.inv(nportmatrix))
        return NPort(self.freqs, inverted, self.type, self.z0)

    def recombine(self, portsets):
        """Recombine ports, reducing the number of ports of this NPort.
        
        :type portsets: a list of ints and tuples
        :param portsets: An int specifies the number of
           a port that needs to be kept as-is. If the int is negative, the
           port's polarity will be reversed. A tuple specifies a pair of ports
           that are to be recombined into one port. The second element of this
           tuple acts as the ground reference to the first element.
        
        >>> recombine([(1,3), (2,4), 5, -6]
        
        will generate a four-port where:
        
        * port 1 is original port 1 referenced to port 3
        * port 2 is original port 2 referenced to port 4
        * port 3 is original port 5
        * port 4 is original port 6, but with reversed polarity

        """
        recombined = []
        for i, matrix in enumerate(self):
            nportmatrix = NPortMatrix(matrix, self.type, self.z0)
            recomb = nportmatrix.convert(IMPEDANCE).recombine(portsets)
            recombined.append(recomb.convert(self.type))
        return NPort(self.freqs, recombined, self.type, self.z0)

    def ispassive(self):
        """Check whether this NPort is passive"""
        for nportmatrix in self:
            if not NPortMatrix(nportmatrix, self.type, self.z0).ispassive():
                return False
        return True


def dot(arg1, arg2):
    """Matrix multiplication for NPorts"""
    def merge_freqs(freqs1, freqs2):
        minf = max(freqs1[0], freqs2[0])
        maxf = min(freqs1[-1], freqs2[-1])
        result = list(set(freqs1).union(set(freqs2)))
        result.sort()
        return np.array(result[result.index(minf):result.index(maxf)])
    
    if type(arg1) == NPort:
        if type(arg2) == NPort:
            result_freqs = merge_freqs(arg1.freqs, arg2.freqs)
            arg1_matrices = arg1.at(result_freqs)
            arg2_matrices = arg2.at(result_freqs)
            result_matrices = np.asarray([np.dot(a, b) for (a,b) in
                zip(arg1_matrices, arg2_matrices)])
        else:
            result_freqs = arg1.freqs
            result_matrices = np.array([np.dot(matrix, other)
                for matrix in arg1.matrices])
        return NPort(result_freqs, result_matrices, arg1.type, arg1.z0)
    elif type(arg1) == twonport.TwoNPort:
        if type(arg2) == twonport.TwoNPort:
            def twonport_dot(l, r):
                a = np.dot(l[0,0], r[0,0]) + np.dot(l[0,1], r[1,0])
                b = np.dot(l[0,0], r[0,1]) + np.dot(l[0,1], r[1,1])
                c = np.dot(l[1,0], r[0,0]) + np.dot(l[1,1], r[1,0])
                d = np.dot(l[1,0], r[0,1]) + np.dot(l[1,1], r[1,1])
                return np.asarray([[a, b], [c, d]])
            
            result_freqs = merge_freqs(arg1.freqs, arg2.freqs)
            arg1_matrices = arg1.at(result_freqs)
            arg2_matrices = arg2.at(result_freqs)
            result_matrices = np.asarray([twonport_dot(a, b) for (a,b) in
                zip(arg1_matrices, arg2_matrices)])
        else:
            raise NotImplementedError
        return twonport.TwoNPort(result_freqs, result_matrices, arg1.type,
                                 arg1.z0)
    elif type(arg2) == NPort:
        raise NotImplementedError
        
        
    raise NotImplementedError



import numpy as np


def parameter(real=None, imag=None, mag=None, db=None, deg=None, rad=None):
    """Initialize a parameter

    Specify:
    * real and (optionally) imag, or
    * mag/db and deg/rad

    """
    if real is not None:
        assert (mag is None) and (db is None) and (deg is None) and (rad is None)
        if imag is None:
            imag = 0
    else: # real is None
        assert imag is None
        if mag is not None:
            assert db is None
        if db is not None:
            assert mag is None
            mag = np.pow(10, db/20.0)
        if rad is not None:
            assert deg is None
        if deg is not None:
            assert rad is None
            rad = np.radians(deg)
        real = mag * np.cos(rad)
        imag = mag * np.sin(rad)
    return complex(real, imag)

def real(arg):
    """Return the real part of the elements of the array"""
    return np.real(arg)

def imag(arg):
    """Return the imaginary part of the elements of the array"""
    return np.imag(arg)

def mag(arg):
    """Return the magnitude of the elements of the array"""
    return np.abs(arg)

def db(arg):
    """Return the magnitude in decibels of the elements of the array"""
    return 20.0 * np.log10(mag(arg))

def rad(arg):
    """Return the phase in radians of the elements of the array"""
    return np.angle(arg)

def deg(arg):
    """Return the phase in degrees of the elements of the array"""
    return np.angle(arg, deg=True)



import touchstone, citi

def main():
    import sys
    from optparse import OptionParser

    usage = "usage: %prog [options] <infile> <outfile (without extension)>"
    parser = OptionParser(usage=usage)
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="be verbose")
    parser.add_option("-f", "--format",
                      action="store", type="string", dest="format",
                      metavar="FORMAT", default="tstone",
                      help="set output format (tstone or citi)")
    parser.add_option("-r", "--recombine", 
                      action="store", type="string", dest="recombine",
                      metavar="PORTLIST", help="recombine ports")

    (options, args) = parser.parse_args()

    if len(args) != 2:
        print("ERROR: Incorrect number of arguments supplied")
        parser.print_help()
        sys.exit(1)

    for format in (touchstone, citi):
        try:
            nport = format.read(args[0], options.verbose)
            break
        except:
            nport = None
    
    if nport is None:
        print("ERROR: Unsupported input file format")
        sys.exit(1)

    if options.recombine:
        nport = nport.recombine(eval("[" + options.recombine + "]"))

    outformats = {'tstone': touchstone, 'citi': citi}
    outformat = outformats[options.format.lower()]
    outformat.write(nport, args[1])




from __future__ import division

import numpy as np
import nport
from twonport import TwoNPort


class TransmissionLine(object):
    """Class representing a two-conductor transmission line. It takes the
    2-port parameters of a transmission line and its length, and it calculates
    the transmission lines':

    * propagation constant :attr:`gamma`, 
    * characteristic impedance :attr:`z0`, and
    * per-unit-length parameters (:attr:`rpm`, :attr:`lpm`, :attr:`gpm`, :attr:`cpm`)

    """
    def __init__(self, twonport, length):
        """
        :param twonport: 2-port that represents a transmission line
        :type twonport: TwoNPort
        :param length: physical length of the transmission line in meters
        :type length: float
        
        """
        assert twonport.ports == 2
        self.twonport = twonport.convert(nport.TRANSMISSION)
        self.freqs = self.twonport.freqs
        self.length = length
        
        # retrieve ABCD parameters
        self.a = self.twonport.get_parameter(1, 1)[:,0,0]
        self.b = self.twonport.get_parameter(1, 2)[:,0,0]
        self.c = self.twonport.get_parameter(2, 1)[:,0,0]
        self.d = self.twonport.get_parameter(2, 2)[:,0,0]

        # the arccosh can be implemented in a number of different ways
        numpy_acosh = np.arccosh

        def mathworld_acosh(z):
            # http://mathworld.wolfram.com/InverseHyperbolicCosine.html
            return np.log(z + np.sqrt(z + 1) * np.sqrt(z - 1))

        def matlab_acosh(z):
            #MATLAB
            return np.log(z + np.sqrt(z*z - 1))
            
        acosh = matlab_acosh

        tmp = (self.a + self.d) / 2.0
        self._gamma = acosh(tmp) / self.length
        self._z0 = np.sqrt(self.b / self.c)

        # TODO: unwrap gamma (*before* division by length)
        
        # extract RLGC parameters [EIS92]
        self._rpm = (self.gamma * self.z0).real
        self._lpm = (self.gamma * self.z0).imag / (2 * np.pi * self.freqs)
        self._gpm = (self.gamma / self.z0).real
        self._cpm = (self.gamma / self.z0).imag / (2 * np.pi * self.freqs)

    @property
    def gamma(self):
        """Propagation constant"""
        return self._gamma

    @property
    def z0(self):
        """Characteristic impedance"""
        return self._z0

    @property
    def rpm(self):
        """Resistance per meter"""
        return self._rpm

    @property
    def lpm(self):
        """Inductance per meter"""
        return self._lpm

    @property
    def gpm(self):
        """Conductance per meter"""
        return self._gpm

    @property
    def cpm(self):
        """Capacitance per meter"""
        return self._cpm



from __future__ import division

import numpy as np
import nport


class TwoNPortMatrix(nport.NPortMatrixBase):
    """Class representing a 2n-port matrix (Z, Y, S, T, G, H or ABCD)"""
    # See also "block matrices" in the Matrix Cookbook
    def __new__(cls, matrix, mtype, z0=None):
        obj = nport.NPortMatrixBase.__new__(cls, matrix, mtype, z0)
        if len(obj.shape) != 4:
            raise ValueError("the matrix should be four-dimensional")
        if obj.shape[0] != 2 or obj.shape[0] != obj.shape[1]:
            raise ValueError("the matrix should be a 2x2 matrix")
        if obj.shape[2] != obj.shape[3]:
            raise ValueError("the submatrices should be square")
        return obj

    @property
    def ports(self):
        return 2 * self.shape[2]

    def nportmatrix(self):
        matrix = np.vstack((np.hstack((self[0,0], self[0,1])),
            np.hstack((self[1,0], self[1,1]))))
        return nport.NPortMatrix(matrix, self.type, self.z0)

    def convert(self, mtype, z0=None):
        """Convert from one 2n-port matrix representation to another"""
        z0 = nport.NPortMatrixBase.__z0test__(mtype, z0)
        identity = np.identity(self.shape[2], dtype=complex)

        # TODO: check for singularities
        if self.type == nport.SCATTERING and mtype == nport.TRANSMISSION:
                s11 = np.asmatrix(self[0,0])
                s12 = np.asmatrix(self[0,1])
                s21 = np.asmatrix(self[1,0])
                s22 = np.asmatrix(self[1,1])
                s21i = np.linalg.inv(s21)
                a = 0.5 * ((identity + s11) * s21i * (identity - s22) + s12)
                b = self.z0 / 2 * ((identity + s11) * s21i * (identity + s22) - s12)
                c = 0.5 / self.z0 * ((identity - s11) * s21i * (identity - s22) - s12)
                d = 0.5 * ((identity - s11) * s21i * (identity + s22) + s12)
                result = [[a, b], [c, d]]
        elif self.type == nport.SCATTERING and mtype == nport.SCATTERING_TRANSFER:
            raise NotImplementedError
        elif self.type == nport.TRANSMISSION and mtype == nport.SCATTERING:
                # see PhDwiki
                a = np.asmatrix(self[0,0])
                b0 = np.asmatrix(self[0,1]) / z0
                c0 = np.asmatrix(self[1,0]) * z0
                d = np.asmatrix(self[1,1])
                x = np.linalg.inv(a + b0 + c0 + d)
                s11 = (a + b0 - c0 -d) * x
                s12 = 0.5 * ((a - b0 - c0 + d) -
                    (a + b0 - c0 - d) * x * (a - b0 + c0 -d))
                s21 = 2 * x
                s22 = x * (-a + b0 - c0 + d)
                result = [[s11, s12], [s21, s22]]
                # alternative formulas [KHA06] p.21
        elif self.type == nport.TRANSMISSION and mtype == nport.SCATTERING_TRANSFER:
            raise NotImplementedError
        else:
            result = self.nportmatrix().convert(mtype, z0)
            return result.twonportmatrix()
            # ABCD -> Z and Y conversions in [PAU08] section 7.5.5

        return TwoNPortMatrix(result, mtype, z0)


class TwoNPort(nport.NPortBase):
    """Class representing a 2n-port across a list of frequencies"""
    def __new__(cls, freqs, matrices, mtype, z0=None):
        obj = nport.NPortBase.__new__(cls, freqs, matrices, mtype, z0)
        if len(obj[0].shape) != 4:
            raise ValueError("the matrices should be four-dimensional")
        if obj[0].shape[0] != 2 or obj[0].shape[0] != obj[0].shape[1]:
            raise ValueError("the matrices should be a 2x2 matrix")
        if obj[0].shape[2] != obj[0].shape[3]:
            raise ValueError("the submatrices should be square")
        return obj

    def __init__(self, freqs, matrices, type, z0=50):
        """
        Initialize an instance, specifying the frequency samples, all
        corresponsing matrices, the matrix type and the reference impedance

        """
        pass

    def __getitem__(self, arg):
        if type(arg) == int:
            return TwoNPortMatrix(np.asarray(self).__getitem__(arg), self.type, self.z0)
        else:
            return np.asarray(self).__getitem__(arg)

    @property
    def ports(self):
        """Return the number of ports of this TwoNPort"""
        return 2 * self[0].shape[2]

    def add(self, freq, matrix):
        """
        Return a TwoNPort with the specified frequency sample added

        If matrix is a TwoNPortMatrix, its elements will be converted to this
        TwoNPort's type and characteristic impedance.
        If matrix is a complex array, it is assumed the elements are in the
        format of this TwoNPort

        """
        if type(matrix) == TwoNPortMatrix:
            if matrix.type != self.type or matrix.z0 != self.z0:
                matrix = matrix.convert(self.type, self.z0)
        index = self.freqs.searchsorted(freq)
        freqs = np.insert(self.freqs, index, freq)
        matrices = np.insert(self.matrices, index, matrix, 0)
        return TwoNPort(freqs, matrices, self.type, self.z0)

    def nport(self):
        """Convert this TwoNPort to an NPort"""
        nportmatrices = []
        for matrix in self:
            twonportmatrix = TwoNPortMatrix(matrix, self.type, self.z0)
            nportmatrices.append(twonportmatrix.nportmatrix())
        return nport.NPort(self.freqs, nportmatrices, self.type, self.z0)

    def convert(self, type, z0=None):
        """Convert to another 2n-port matrix representation"""
        converted = []
        for matrix in self:
            twonportmatrix = TwoNPortMatrix(matrix, self.type, self.z0)
            converted.append(twonportmatrix.convert(type, z0))
        return TwoNPort(self.freqs, converted, type, z0)

    #~ def invert(self):
        #~ """Return a TwoNPort described by the inverse of this TwoNPort's
        #~ matrices
        
        #~ """
        #~ inverted = []
        #~ for matrix in self:
            #~ nportmatrix = NPortMatrix(matrix, self.type, self.z0)
            #~ inverted.append(np.linalg.inv(nportmatrix))
        #~ return NPort(self.freqs, inverted, self.type, self.z0)



#!/bin/env python

from setuptools import setup
from subprocess import Popen, PIPE

# write the git version to nport/version.py
# based on version.py by Douglas Creager <dcreager@dcreager.net>
# http://dcreager.net/2010/02/10/setuptools-git-version-numbers/
try:
    p = Popen(['git', 'describe', '--abbrev=4'],
              stdout=PIPE, stderr=PIPE)
    p.stderr.close()
    line = p.stdout.readlines()[0]
    version = line.strip()[1:]
except:
    print("A problem occured while trying to run git. "
          "Version information is unavailable!")
    version = 'unknown'

version_file = open('nport/version.py', 'w')
version_file.write("__version__ = '%s'\n" % version)
version_file.close()


setup(
    name='nport',
    version=version,
    packages=['nport'],
    scripts=['nporttool'],
    requires=['numpy','scipy'],
    
    author="Brecht Machiels",
    author_email="brecht.machiels@esat.kuleuven.be",
    description="Python package for handling n-port data",
    license="GPL",
    keywords="two-port 2n-port s-parameters touchstone citi deembedding",
    url="https://github.com/bmachiel/python-nport",
)
